/*
 *	Low-Level PCI Support for PC
 *
 *	(c) 1999--2000 Martin Mares <mj@ucw.cz>
 */

/* From file arch/i386/kernel/pci-pc.c */

#include <linux/module.h>
#include <linux/pci.h>

#define VENDOR_ID	0x10EC
#define DEVICE_ID	0x8139
#define CLASS_CODE	0x00020000	/* Network controller */

#define PCIBIOS_PCI_BIOS_PRESENT 	0xb101
#define PCIBIOS_FIND_PCI_DEVICE		0xb102
#define PCIBIOS_FIND_PCI_CLASS_CODE	0xb103
#define PCIBIOS_READ_CONFIG_BYTE	0xb108
#define PCIBIOS_READ_CONFIG_WORD	0xb109
#define PCIBIOS_READ_CONFIG_DWORD	0xb10a

/* BIOS32 signature: "_32_" */
#define BIOS32_SIGNATURE	(('_' << 0) + ('3' << 8) + ('2' << 16) + ('_' << 24))

/* PCI signature: "PCI " */
#define PCI_SIGNATURE		(('P' << 0) + ('C' << 8) + ('I' << 16) + (' ' << 24))

/* PCI service signature: "$PCI" */
#define PCI_SERVICE		(('$' << 0) + ('P' << 8) + ('C' << 16) + ('I' << 24))

struct pci_dev_struct {
    u16 vendor_id, device_id;
    u32 class_code, base_addr;
    u8 bus, dev, fn;
};


/*
 * Physical address of the service directory.  I don't know if we're
 * allowed to have more than one of these or not, so just in case
 * we'll make pcibios_present() take a memory start parameter and store
 * the array there.
 */

static struct {
	u32 address;
	u16 segment;
} bios32_indirect = { 0, __KERNEL_CS };


static struct {
	u32 address;
	u16 segment;
} pci_indirect = { 0, __KERNEL_CS };



/*
 * This is the standard structure used to identify the entry point
 * to the BIOS32 Service Directory, as documented in
 * 	Standard BIOS 32-bit Service Directory Proposal
 * 	Revision 0.4 May 24, 1993
 * 	Phoenix Technologies Ltd.
 *	Norwood, MA
 * and the PCI BIOS specification.
 */

union bios32 {
	struct {
		u32 signature;	/* _32_ */
		u32 entry;	/* 32 bit physical address */
		u8 revision;	/* Revision level, 0 */
		u8 length;	/* Length in paragraphs should be 01 */
		u8 checksum;	/* All bytes must add up to zero */
		u8 reserved[5];	/* Must be zero */
	} fields;
	char chars[16];
};


/*
 * Try to find PCI BIOS.
 */

int pci_find_bios(void)
{
	union bios32 *check;
	u8 sum;
	int i, length;

	/*
	 * Follow the standard procedure for locating the BIOS32 Service
	 * directory by scanning the permissible address range from
	 * 0xe0000 through 0xfffff for a valid BIOS32 structure.
	 */

	for (check = (union bios32 *) __va(0xe0000);
	     check <= (union bios32 *) __va(0xffff0);
	     ++check) {
		if (check->fields.signature != BIOS32_SIGNATURE)
			continue;
		length = check->fields.length * 16;
		if (!length)
			continue;
		sum = 0;
		for (i = 0; i < length ; ++i)
			sum += check->chars[i];
		if (sum != 0)
			continue;
		if (check->fields.revision != 0) {
			printk("PCI: unsupported BIOS32 revision %d at 0x%p\n",
				check->fields.revision, check);
			continue;
		}
		if (check->fields.entry >= 0x100000) {
			printk("PCI: BIOS32 entry (0x%p) in high memory, cannot use.\n", check);
			return 0;
		} else {
			unsigned long bios32_entry = check->fields.entry;
			bios32_indirect.address = bios32_entry + PAGE_OFFSET;
			printk(KERN_INFO "PCI: BIOS32 entry point at 0x%08x\n", bios32_indirect.address);
		}
		break;	/* Hopefully more than one BIOS32 cannot happen... */
	}

	return 0;
}


static u32 bios32_service(u32 service)
{
	u8 return_code;		/* %al */
	u32 address;		/* %ebx */
	u32 length;		/* %ecx */
	u32 entry;		/* %edx */
	u32 flags;

	__save_flags(flags); __cli();
	__asm__("lcall (%%edi); cld"
		: "=a" (return_code),
		  "=b" (address),
		  "=c" (length),
		  "=d" (entry)
		: "0" (service),
		  "1" (0),
		  "D" (&bios32_indirect));
	__restore_flags(flags);

	switch (return_code) {
		case 0:
			return address + entry;
		case 0x80:	/* Not present */
			printk(KERN_WARNING "bios32_service(0x%08x): not present\n", service);
			return 0;
		default: /* Shouldn't happen */
			printk(KERN_WARNING "bios32_service(0x%08x): returned 0x%x -- BIOS bug!\n",
				service, return_code);
			return 0;
	}
}


int check_pcibios(void)
{
	u32 signature, eax, ebx, ecx;
	u8 status, major_ver, minor_ver, hw_mech;
	u32 flags, pcibios_entry;

	if ((pcibios_entry = bios32_service(PCI_SERVICE))) {
		pci_indirect.address = pcibios_entry + PAGE_OFFSET;

		__save_flags(flags); __cli();
		__asm__(
			"lcall (%%edi); cld\n\t"
			"jc 1f\n\t"
			"xor %%ah, %%ah\n"
			"1:"
			: "=d" (signature),
			  "=a" (eax),
			  "=b" (ebx),
			  "=c" (ecx)
			: "1" (PCIBIOS_PCI_BIOS_PRESENT),
			  "D" (&pci_indirect)
			: "memory");
		__restore_flags(flags);

		status = (eax >> 8) & 0xff;
		hw_mech = eax & 0xff;
		major_ver = (ebx >> 8) & 0xff;
		minor_ver = ebx & 0xff;

		if (status || signature != PCI_SIGNATURE) {
			printk (KERN_ERR "PCI: BIOS BUG #%x[%08x] found\n",
				status, signature);
			return 0;
		}
		printk(KERN_INFO "PCI: PCI BIOS revision %x.%02x entry at 0x%08x\n",
			major_ver, minor_ver, pcibios_entry);

		return 1;
	}
	return 0;
}




static int pci_bios_find_device(u16 vendor, u16 device_id, u16 index, u8 *bus, u8 *dev, u8 *fn)
{
	u16 bx;
	u16 ret;

	__asm__("lcall (%%edi); cld\n\t"
		"jc 1f\n\t"
		"xor %%ah, %%ah\n"
		"1:"
		: "=b" (bx),
		  "=a" (ret)
		: "1" (PCIBIOS_FIND_PCI_DEVICE),
		  "c" (device_id),
		  "d" (vendor),
		  "S" ((int) index),
		  "D" (&pci_indirect));

	*bus = (bx >> 8) & 0xff;
	*dev = (bx & 0xff) >> 3;
	*fn = bx & 0x3;

	return (int) (ret & 0xff00) >> 8;
}



static int pci_bios_find_class(u32 class_code, u16 index, struct pci_dev_struct *pd)
{
	u16 bx;
	u16 ret;

	__asm__("lcall (%%edi); cld\n\t"
		"jc 1f\n\t"
		"xor %%ah, %%ah\n"
		"1:"
		: "=b" (bx),
		  "=a" (ret)
		: "1" (PCIBIOS_FIND_PCI_CLASS_CODE),
		  "c" (class_code),
		  "S" ((int) index),
		  "D" (&pci_indirect));

	pd->bus = (bx >> 8) & 0xff;
	pd->dev = (bx & 0xff) >> 3;
	pd->fn = bx & 0x03;
	return (int) (ret & 0xff00) >> 8;
}



static int pci_bios_read(int bus, int dev, int fn, int reg, int len, u32 *value)
{
	u32 result = 0;
	u32 bx = ((bus << 8) | (dev << 3) | fn);

	if (bus > 255 || dev > 31 || fn > 7 || reg > 255)
		return -EINVAL;

	switch (len) {
	case 1:
		__asm__("lcall (%%esi); cld\n\t"
			"jc 1f\n\t"
			"xor %%ah, %%ah\n"
			"1:"
			: "=c" (*value),
			  "=a" (result)
			: "1" (PCIBIOS_READ_CONFIG_BYTE),
			  "b" (bx),
			  "D" ((long)reg),
			  "S" (&pci_indirect));
		break;
	case 2:
		__asm__("lcall (%%esi); cld\n\t"
			"jc 1f\n\t"
			"xor %%ah, %%ah\n"
			"1:"
			: "=c" (*value),
			  "=a" (result)
			: "1" (PCIBIOS_READ_CONFIG_WORD),
			  "b" (bx),
			  "D" ((long)reg),
			  "S" (&pci_indirect));
		break;
	case 4:
		__asm__("lcall (%%esi); cld\n\t"
			"jc 1f\n\t"
			"xor %%ah, %%ah\n"
			"1:"
			: "=c" (*value),
			  "=a" (result)
			: "1" (PCIBIOS_READ_CONFIG_DWORD),
			  "b" (bx),
			  "D" ((long)reg),
			  "S" (&pci_indirect));
		break;
	}

	return (int)((result & 0xff00) >> 8);
}




/* Read PCI direct */
#define PCI_CONF1_ADDRESS(bus, dev, fn, reg) \
	(0x80000000 | (bus << 16) | (dev << 11) | (fn << 8) | (reg & ~3))

static int pci_direct_read(int bus, int dev, int fn, int reg, int len, u32 *value)
{
	if (bus > 255 || dev > 31 || fn > 7 || reg > 255)
		return -EINVAL;

	outl(PCI_CONF1_ADDRESS(bus, dev, fn, reg), 0xCF8);

	switch (len) {
	case 1:
		*value = inb(0xCFC + (reg & 3));
		break;
	case 2:
		*value = inw(0xCFC + (reg & 2));
		break;
	case 4:
		*value = inl(0xCFC);
		break;
	}

	return 0;
}




int pci_direct_find_class(u32 class_code, struct pci_dev_struct *pd)
{
    int bus, dev, fn = 0, idx = 0x08;
    u32 config_dword, code;

    printk(KERN_INFO "Looking for device with class code 0x%X\n", class_code);

    memset(pd, 0, sizeof(struct pci_dev_struct));

/*
 * bus - 0 .. 255
 * dev - 0 .. 31
 * fn - 0 .. 7
 */

    for(bus = 0; bus < 256; bus++) {
	for(dev = 0; dev < 32; dev++) {
	    for(fn = 0; fn < 8; fn++) {

	        pci_direct_read(bus, dev, fn, idx, 4, &config_dword);
		code = config_dword >> 8;

		if(code == class_code) {

		    printk(KERN_INFO "OK. Device find.\n");
		    printk(KERN_INFO "bus - %d, dev - %d, fn - %d\n", bus, dev, fn);
		    printk(KERN_INFO "Class Code - 0x%x\n", code);

		    pd->bus = bus;
		    pd->dev = dev;
		    pd->fn = fn;

		    return 0;
		}
	    }
	}
    }

    return 0x80;
}



void get_mac_addr(u32 base_addr)
{
    int i = 0;
    u8 mac[6];

    memset(mac, 0, 6);

/* Get and display MAC address */
    for(; i < 6; i++)
        mac[i] = inb(base_addr + i);

    printk(KERN_INFO "MAC address: %02X:%02X:%02X:%02X:%02X:%02X\n",
	    mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

    return;
}


void display_pcidev_info(struct pci_dev_struct *pdev)
{
    printk(KERN_INFO "VENDOR ID - 0x%X\n", pdev->vendor_id);
    printk(KERN_INFO "DEVICE ID - 0x%X\n", pdev->device_id);
    printk(KERN_INFO "CLASS CODE - 0x%X\n", pdev->class_code);
    printk(KERN_INFO "BASE ADDRESS - 0x%X\n", pdev->base_addr);

    return;
}


static int __init pcidev_on(void)
{
    struct pci_dev_struct pdev;
    int idx = 0;
    u8 bus = 0, dev = 0, fn = 0;
    u16 command_reg = 0;
    u32 config_dword = 0;

    pci_find_bios();
    check_pcibios();


    if(pci_bios_find_device(VENDOR_ID, DEVICE_ID, idx, &bus, &dev, &fn) == PCIBIOS_SUCCESSFUL)
	printk(KERN_INFO "Device found by type, bus - %d, dev - %d, fn - %d\n", bus, dev, fn);


    memset((void *)&pdev, 0, sizeof(struct pci_dev_struct));

    if(pci_bios_find_class(CLASS_CODE, idx, &pdev) == PCIBIOS_SUCCESSFUL)
	printk(KERN_INFO "Device found by class, bus - %d, dev - %d, fn - %d\n",
	    pdev.bus, pdev.dev, pdev.fn);
	else {
	    printk(KERN_INFO "Device not founf by class\n");
	    return 0;
	}


/* Read VENDOR ID */
    idx = 0x00;
    if(pci_bios_read(pdev.bus, pdev.dev, pdev.fn, idx, 2, &config_dword) == PCIBIOS_SUCCESSFUL)
	pdev.vendor_id = (u16)config_dword;


/* Read DEVICE ID */
    idx = 0x02;
    if(pci_bios_read(pdev.bus, pdev.dev, pdev.fn, idx, 2, &config_dword) == PCIBIOS_SUCCESSFUL)
	pdev.device_id = (u16)config_dword;


/* Read Class Code */
    idx = 0x08;
    if(pci_bios_read(pdev.bus, pdev.dev, pdev.fn, idx, 4, &config_dword) == PCIBIOS_SUCCESSFUL)
	pdev.class_code = config_dword >> 8;


/* Read Command Register */
    idx = 0x04;
    if(pci_bios_read(pdev.bus, pdev.dev, pdev.fn, idx, 1, &config_dword) == PCIBIOS_SUCCESSFUL)
	command_reg = config_dword;


/* Read Base Address Registers */
    idx = 0x10;
    if(command_reg & 0x01) {
	for(; idx < 0x28 ;) {
	    if(pci_bios_read(pdev.bus, pdev.dev, pdev.fn, idx, 4, &config_dword) == PCIBIOS_SUCCESSFUL) {
		if(config_dword & 0x01) {
		    config_dword &= ~0x1;
		    pdev.base_addr = config_dword;
		    break;
		}
		idx += 4;
	    }
	}
    } else return 0;

    display_pcidev_info(&pdev);
    get_mac_addr(pdev.base_addr);

/* Direct read PCI */
    printk(KERN_INFO "PCI direct access:\n");

    if(pci_direct_find_class(CLASS_CODE, &pdev) < 0) {
	printk(KERN_INFO "Device not found\n");
	return 0;
    }

/* Read VENDOR_ID */
    idx = 0x00;
    if(pci_direct_read(pdev.bus, pdev.dev, pdev.fn, idx, 2, &config_dword) == 0)
        printk(KERN_INFO "VENDOR_ID - 0x%X\n", config_dword);

/* Read DEVICE_ID */
    idx = 0x02;
    if(pci_direct_read(pdev.bus, pdev.dev, pdev.fn, idx, 2, &config_dword) == 0)
        printk(KERN_INFO "DEVICE_ID - 0x%X\n", config_dword);

/* Read Class Code */
    idx = 0x08;
    if(pci_direct_read(pdev.bus, pdev.dev, pdev.fn, idx, 4, &config_dword) == 0) {
        printk(KERN_INFO "CLASS_CODE - 0x%X\n", config_dword >> 8);
    }

/* Read Command Register */
    command_reg = 0;
    idx = 0x04;
    if(pci_direct_read(pdev.bus, pdev.dev, pdev.fn, idx, 2, &config_dword) == 0)
	command_reg = config_dword;


/* Read Base Address Registers */
    idx = 0x10;
    if(command_reg & 0x01) {
	for(; idx < 0x28 ;) {
	    if(pci_direct_read(pdev.bus, pdev.dev, pdev.fn, idx, 4, &config_dword) == PCIBIOS_SUCCESSFUL) {
		if(config_dword & 0x01) {
		    config_dword &= ~0x1;
		    pdev.base_addr = config_dword;
		    break;
		}
		idx += 4;
	    }
	}
    } else return 0;

    get_mac_addr(config_dword);

    return 0;
}



static void __exit pcidev_off(void)
{
    return;
}


module_init(pcidev_on);
module_exit(pcidev_off);
